from ..helpers import ResourceBase, FilteredIterableResource
from ..errors import ok_or_error, response_or_error
from ..compat import update_doc

class Users(ResourceBase, FilteredIterableResource):
    @response_or_error
    def add(self, name, password, displayName, emailAddress, addToDefaultGroup=True):
        """
        Add a user, returns a dictionary containing information about the newly created user
        """
        params = dict(name=name,
                    password=password,
                    displayName=displayName,
                    emailAddress=emailAddress,
                    addToDefaultGroup=addToDefaultGroup)

        return self._client.post(self.url(), params=params)

    @ok_or_error
    def delete(self, user):
        """
        Delete a user.
        """
        return self._client.delete(self.url(), params=dict(name=user))

    @response_or_error
    def update(self, name, displayName=None, emailAddress=None):
        """
        Update the user information, and return the updated user info.

        None is used as a sentinel value, use empty string if you mean to clear.
        """

        data = dict(name=name)
        if displayName is not None:
            data['displayName'] = displayName
        if data is not None:
            data['emailAddress'] = emailAddress

        return self._client.put(self.url(), data)

    @ok_or_error
    def credentials(self, name, new_password):
        """
        Update a user's password.
        """

        data = dict(name=name, password=new_password, passwordConfirm=new_password)
        return self._client.put(self.url(), data)

    @ok_or_error
    def add_group(self, user, group):
        """
        Add the given group to the given user.
        """
        return self._client.post(self.url("/add-group"), dict(context=user, itemName=group))

    @ok_or_error
    def remove_group(self, user, group):
        """
        Remove the given group from the given user.
        """
        return self._client.post(self.url("/remove-group"), dict(context=user, itemName=group))

    def more_members(self, user, filter=None):
        """
        Retrieves a list of groups the specified user is a member of.

        filter: if specified only groups with names containing the supplied string will be returned
        """
        params = dict(context=user)
        if filter:
            params['filter'] = filter
        return self.paginate("/more-members", params)

    def more_non_members(self, user, filter=None):
        """
        Retrieves a list of groups that the specified user is not a member of

        filter: if specified only groups with names containing the supplied string will be returned
        """
        params = dict(context=user)
        if filter:
            params['filter'] = filter
        return self.paginate("/more-non-members", params)


update_doc(Users.all, """
Returns an iterator that will walk all the users, paginating as necessary.

filter: return only users with usernames, display name or email addresses containing the supplied string
""")



class FileInfo:
    def __init__(self, file_info):
        self.components = file_info["components"]
        self.parent = file_info["parent"]
        self.name = file_info["name"]
        self.extension = file_info.get("extension","")
        self.toString = file_info["toString"]

    def _get_components(self):
        return self._components

    def _set_components(self, value):
        self._components = value

    def _get_parent(self):
        return self._parent

    def _set_parent(self, value):
        self._parent = value

    def _get_name(self):
        return self._name

    def _set_name(self, value):
        self._name = value

    def _get_extension(self):
        return self._extension

    def _set_extension(self, value):
        self._extension = value

    def _get_to_string(self):
        return self._toString

    def _set_to_string(self, value):
        self._toString = value

    components = property(_get_components, _set_components, doc="The components the file reside in.")

    parent = property(_get_parent, _set_parent, doc="The parent folder the file resides in.")

    name = property(_get_name, _set_name, doc="The name of the file.")

    extension = property(_get_extension, _set_extension, doc="The extension of the file.")

    toString = property(_get_to_string, _set_to_string,
                        doc="A string value representing the full path from repository root.")



from ..helpers import Nested, ResourceBase

from .groups import Groups
from .users import Users
from ..permissions import Permissions


class Admin(ResourceBase):
    groups = Nested(Groups)
    users = Nested(Users)
    permissions = Nested(Permissions)



from .helpers import Nested, ResourceBase, IterableResource
from .permissions import ProjectPermissions
from .errors import ok_or_error, response_or_error
from .repos import Repos
from .compat import update_doc


class Project(ResourceBase):
    def __init__(self, key, url, client, parent):
        super(Project, self).__init__(url, client, parent)
        self._key = key

    @ok_or_error
    def delete(self):
        """
        Delete the project
        """
        return self._client.delete(self.url())

    @response_or_error
    def update(self, new_key=None, name=None, description=None, avatar=None, public=None):
        """
        Update project information. If supplied, avatar should be a base64 encoded image.

        None is used as a sentinel so use '' to clear a value.
        """
        data = dict()
        if new_key is not None:
            data['key'] = new_key
        if name is not None:
            data['name'] = name
        if description is not None:
            data['description'] = description
        if avatar is not None:
            data['avatar'] = "data:image/png;base64," + avatar
        if public is not None:
            data['public'] = public

        return self._client.put(self.url(), data)

    @response_or_error
    def get(self):
        return self._client.get(self.url())

    permissions = Nested(ProjectPermissions, relative_path="/permissions")
    repos = Nested(Repos)


class Projects(ResourceBase, IterableResource):
    @response_or_error
    def get(self, project):
        """
        Retrieve the project matching the supplied key.
        """
        return self._client.get(self.url(project))

    def __getitem__(self, item):
        return Project(item, self.url(item), self._client, self)


    @response_or_error
    def create(self, key, name, description='', avatar=None):
        """
        Create a project. If supplied, avatar should be a base64 encoded image.
        """
        data = dict(key=key, name=name, description=description)
        if avatar:
            data['avatar'] = "data:image/png;base64," + avatar
        return self._client.post(self.url(), data)


update_doc(Projects.all, """Retrieve projects.""")



from .diffs import Diff
from .helpers import ResourceBase
from .errors import response_or_error


class PullRequestDiffRef(object):
    def __init__(self, project_key, repo_slug, id):
        self.project_key = project_key
        self.repo_slug = repo_slug
        self.id = id

    def to_dict(self):
        return dict(id=self.id,
                    repository=dict(slug=self.repo_slug,
                                    project=dict(key=self.project_key),
                                    name=self.repo_slug))


class PullRequestDiff(ResourceBase):
    def __init__(self, url, client, parent):
        super(PullRequestDiff, self).__init__(url, client, parent)
        response = self.get()
        self.from_hash = response["fromHash"]
        self.to_hash = response["toHash"]
        self.context_lines = response["contextLines"]
        self.whitespace = response["whitespace"]
        self.diffs = []
        for value in list(response["diffs"]):
            self.diffs.append(Diff(value))

    @response_or_error
    def get(self):
        return self._client.get(self.url())

    def _get_from_hash(self):
        return self._from_hash

    def _set_from_hash(self, value):
        self._from_hash = value

    def _get_to_hash(self):
        return self._to_hash

    def _set_to_hash(self, value):
        self._to_hash = value

    def _get_diffs(self):
        return self._diffs

    def _set_diffs(self, value):
        self._diffs = value

    def _get_context_lines(self):
        return self._context_lines

    def _set_context_lines(self, value):
        self._context_lines = value

    def _get_whitespace(self):
        return self._whitespace

    def _set_whitespace(self, value):
        self._whitespace = value

    from_hash = property(_get_from_hash, _set_from_hash)

    to_hash = property(_get_to_hash, _set_to_hash)

    diffs = property(_get_diffs, _set_diffs)

    context_lines = property(_get_context_lines, _set_context_lines)

    whitespace = property(_get_whitespace, _set_whitespace)



from .helpers import ResourceBase, IterableResource, Nested
from .errors import ok_or_error, response_or_error
from .compat import update_doc

API_NAME = 'branch-permissions'
API_VERSION = '1.0'
API_OVERRIDE_PATH = '{0}/{1}'.format(API_NAME, API_VERSION)

class Permitted(ResourceBase, IterableResource):
    """Get-only resource that describes a permission record"""
    def __init__(self, url, client, parent):
        ResourceBase.__init__(self, url, client, parent, API_OVERRIDE_PATH)

update_doc(Permitted.all, """Retrieve list of permitted entities for a repo""")


class Restriction(ResourceBase):
    def __init__(self, id, url, client, parent):
        super(Restriction, self).__init__(url, client, parent, API_OVERRIDE_PATH)
        self._id = id

    @response_or_error
    def get(self):
        """
        Retrieve a restriction
        """
        return self._client.get(self.url())

    @ok_or_error
    def delete(self):
        """
        Delete a restriction
        """
        return self._client.delete(self.url())

    @response_or_error
    def update(self, value, users=None, groups=None, pattern=False):
        """
        Re-restrict a branch, or set of branches defined by a pattern, to a set of users and/or groups
        """
        data = dict(type=('PATTERN' if pattern else 'BRANCH'), value=value)

        if users is not None:
            data['users'] = users
        if groups is not None:
            data['groups'] = groups

        return self._client.put(self.url(""), data=data)


class Restricted(ResourceBase, IterableResource):

    def __init__(self, url, client, parent):
        ResourceBase.__init__(self, url, client, parent, API_OVERRIDE_PATH)

    def __getitem__(self, item):
        return Restriction(item, self.url(item), self._client, self)

    @response_or_error
    def create(self, value, users=None, groups=None, pattern=False):
        """
        Restrict a branch, or set of branches defined by a pattern, to a set of users and/or groups
        """
        data = dict(type=('PATTERN' if pattern else 'BRANCH'), value=value)

        if users is not None:
            data['users'] = users
        if groups is not None:
            data['groups'] = groups

        return self._client.post(self.url(""), data=data)

update_doc(Restricted.all, """Retrieve list of restrictions for a repo""")

class BranchPermissions(ResourceBase):
    """Simple parent resource for this api, to distinguish permissions from permitted"""
    permitted = Nested(Permitted)
    restricted = Nested(Restricted)



from ..helpers import ResourceBase, FilteredIterableResource
from ..errors import ok_or_error, response_or_error
from ..compat import update_doc

class Groups(ResourceBase, FilteredIterableResource):
    @response_or_error
    def add(self, group):
        """
        Add a group, returns a dictionary containing the group name
        """
        return self._client.post(self.url(), {}, params=dict(name=group))

    @ok_or_error
    def delete(self, group):
        """
        Delete a group.
        """
        return self._client.delete(self.url(), params=dict(name=group))

    @ok_or_error
    def add_user(self, group, user):
        """
        Add a user to a group.
        """
        return self._client.post(self.url("/add-user"), dict(context=group, itemName=user))

    @ok_or_error
    def remove_user(self, group, user):
        """
        Remove a user to a group.
        """
        return self._client.post(self.url("/remove-user"), dict(context=group, itemName=user))

    def more_members(self, group, filter=None):
        """
        Retrieves a list of users that are members of a specified group.

        filter: return only users with usernames, display names or email addresses containing this string
        """
        params = dict(context=group)
        if filter:
            params['filter'] = filter
        return self.paginate("/more-members", params)

    def more_non_members(self, group, filter=None):
        """
        Retrieves a list of users that are not members of a specified group.

        filter: return only users with usernames, display names or email addresses containing this string
        """
        params = dict(context=group)
        if filter:
            params['filter'] = filter
        return self.paginate("/more-non-members", params)


update_doc(Groups.all, """
Returns an iterator that will walk all the groups, paginating as necessary.

filter: if specified only group names containing the supplied string will be returned
""")



from .fileinfo import FileInfo


class Diff:
    def __init__(self, diff_file):
        source_detail = diff_file.get("source", None)
        if source_detail is not None:
            self.source = FileInfo(diff_file["source"])
        else:
            self.source = None

        dest_detail = diff_file.get("destination", None)
        if dest_detail is not None:
            self.destination = FileInfo(diff_file["destination"])
        else:
            self.destination = None

        self.hunks = diff_file.get("hunks", [])
        self.truncated = diff_file.get("truncated", [])
        self.line_comments = diff_file.get("lineComments", [])

    def _get_source(self):
        return self._source

    def _set_source(self, value):
        self._source = value

    def _get_destination(self):
        return self._destination

    def _set_destination(self, value):
        self._destination = value

    def _get_hunks(self):
        return self._hunks

    def _set_hunks(self, value):
        self._hunks = value

    def _get_truncated(self):
        return self._truncated

    def _set_truncated(self, value):
        self._truncated = value

    def _get_line_comments(self):
        return self._line_comments

    def _set_line_comments(self, value):
        self._line_comments = value

    source = property(_get_source, _set_source, doc="The source of a file in the diff.")

    destination = property(_get_destination, _set_destination, doc="The destination of a file in the diff.")

    hunks = property(_get_hunks, _set_hunks, doc="A dictionary showing the hunks changed in the file difference.")

    truncated = property(_get_truncated, _set_truncated,
                         doc="Whether it has been truncated? I'm not actually sure what this value represents from the JSON response.")

    line_comments = property(_get_line_comments, _set_line_comments,
                             doc="The comments that have been made against a file in the pull request.")



from .helpers import ResourceBase, IterableResource
from .errors import ok_or_error, response_or_error
from .compat import basestring
from .pullrequestdiffs import PullRequestDiff
import json


class PullRequestRef(object):
    def __init__(self, project_key, repo_slug, id):
        self.project_key = project_key
        self.repo_slug = repo_slug
        self.id = id

    def to_dict(self):
        return dict(id=self.id,
                    repository=dict(slug=self.repo_slug,
                                    project=dict(key=self.project_key),
                                    name=self.repo_slug))


class PullRequest(ResourceBase):
    def __init__(self, id, url, client, parent):
        super(PullRequest, self).__init__(url, client, parent)
        self._id = id

    @response_or_error
    def get(self):
        """
        Retrieve a pull request.
        """
        return self._client.get(self.url())


    def _make_ref(self, ref, refName):
        if isinstance(ref, basestring):
            repo = self.get()[refName]['repository']
            return PullRequestRef(repo['project']['key'], repo['slug'], ref).to_dict()
        elif isinstance(ref, PullRequestRef):
            return ref.to_dict()
        elif isinstance(ref, dict):
            return ref
        else:
            raise ValueError(refName + " should be either a string, a dict, or a PullRequestRef")


    @response_or_error
    def update(self, version, title=None, description=None, reviewers=None, toRef=None, fromRef=None):
        """
        Update the title, description, references or reviewers of an existing pull request.

        Note: the reviewers list may be updated using this resource. However the author and participants list may not.
        toRef and fromRef might be either strings (e.g. refs/heads/master) or PullRequestRef objects.
        If you want to change the repository of the ref, you need to use a PullRequestRef object.
        """
        data = dict(id=self._id, version=version)
        if title is not None:
            data['title'] = title
        if description is not None:
            data['description'] = description
        if reviewers is not None:
            data['reviewers'] = []
            for reviewer in reviewers:
                data['reviewers'].append({"user": dict(name=reviewer)})
        if toRef is not None:
            data['toRef'] = self._make_ref(toRef, "toRef")
        if fromRef is not None:
            data['fromRef'] = self._make_ref(fromRef, "fromRef")
        return self._client.put(self.url(), data=data)

    def activities(self, fromId=None, fromType=None):
        """
        Retrieve a page of activity associated with a pull request.

        Activity items include comments, approvals, rescopes (i.e. adding and removing of commits), merges and more.

        Different types of activity items may be introduced in newer versions of Stash or by user installed plugins,
        so clients should be flexible enough to handle unexpected entity shapes in the returned page.

        fromId: (optional) the id of the activity item to use as the first item in the returned page
        fromType: (required if fromId is present) the type of the activity item specified by fromId
        """
        params = dict()
        if fromId is not None:
            if fromType is None:
                raise ValueError("fromType is required when fromId is supplied")
            params['fromId'] = fromId
            params['fromType'] = fromType
        return self.paginate("/activities", params=params)

    @ok_or_error
    def decline(self, version=-1):
        """Decline a pull request."""
        return self._client.post(self.url("/decline"), data=dict(version=version))

    def can_merge(self):
        """
        Test whether a pull request can be merged.

        A pull request may not be merged if:

            * there are conflicts that need to be manually resolved before merging; and/or
            * one or more merge checks have vetoed the merge.
        """
        res = self.merge_info()
        return res['canMerge'] and not res['conflicted']

    @response_or_error
    def merge_info(self):
        """
        Show conflicts and vetoes of pull request.

        A pull request may not be merged if:

            * there are conflicts that need to be manually resolved before merging; and/or
            * one or more merge checks have vetoed the merge.
        """
        return self._client.get(self.url("/merge"))

    @response_or_error
    def merge(self, version=-1):
        """
        Merge the specified pull request.
        """
        return self._client.post(self.url("/merge"), data=dict(version=version))

    @response_or_error
    def reopen(self, version=-1):
        """
        Re-open a declined pull request.
        """
        return self._client.post(self.url("/reopen"), data=dict(version=version))

    @response_or_error
    def approve(self):
        """
        Approve a pull request as the current user. Implicitly adds the user as a participant if they are not already.
        """
        data = dict(approved=True, status="approved")
        return self._client.put(self.url("/participants/%s/" % self._client._session.auth[0]), data=data)

    @response_or_error
    def unapprove(self):
        """
        Remove approval from a pull request as the current user. This does not remove the user as a participant.
        """
        data = dict(approved=False, status="unapproved")
        return self._client.put(self.url("/participants/%s/" % self._client._session.auth[0]), data=data)

    @response_or_error
    def watch(self):
        """
        Add the current user as a watcher for the pull request.
        """
        return self._client.post(self.url("/watch"))

    @response_or_error
    def unwatch(self):
        """
        Remove the current user as a watcher for the pull request.
        """
        return self._client.delete(self.url("/watch"))

    def changes(self):
        """
        Gets changes for the specified PullRequest.

        Note: This resource is currently not paged. The server will return at most one page.
        The server will truncate the number of changes to an internal maximum.
        """
        return self.paginate("/changes")

    def commits(self):
        """
        Retrieve changesets for the specified pull request.
        """
        return self.paginate('/commits')

    def comments(self, srcPath='/'):
        """
        Retrieve comments for the specified file in a  pull request.
        """
        return self.paginate('/comments?path=%s' % srcPath)

    @ok_or_error
    def comment(self, commentText, parentCommentId=-1, srcPath=None, fileLine=-1, lineType="CONTEXT", fileType="FROM"):
        """
        Comment on a pull request. If parentCommentId is supplied, it the comment will be
        a child comment of the comment with the id parentCommentId.

        Note: see https://developer.atlassian.com/static/rest/stash/3.11.3/stash-rest.html#idp1448560
        srcPath: (optional) The path of the file to comment on.
        fileLine: (optional) The line of the file to comment on.
        lineType: (optional, defaults to CONTEXT) the type of chunk that is getting commented on.
            Either ADDED, REMOVED, or CONTEXT
        fileType: (optional, defaults to FROM) the version of the file to comment on.
            Either FROM, or TO
        """
        data = dict(text=commentText)
        if parentCommentId is not -1:
            data['parent'] = dict(id=parentCommentId)
        elif srcPath is not None:
            data['anchor'] = dict(path=srcPath, srcPath=srcPath)
            if fileLine is not -1:
                data['anchor'].update(dict(line=fileLine, lineType=lineType, fileType=fileType))
        return self._client.post(self.url("/comments"), data=data)

    def diff(self):
        """
        Retrieve the diff for the specified pull request.
        """
        return PullRequestDiff(self.url('/diff'), self._client, self)


class PullRequests(ResourceBase, IterableResource):
    def __init__(self, url, client, parent):
        super(PullRequests, self).__init__(url, client, parent)

    def all(self, direction='INCOMING', at=None, state='OPEN', order=None, author=None):
        """
        Retrieve pull requests to or from the specified repository.

        direction: (optional, defaults to INCOMING) the direction relative to the specified repository.
            Either INCOMING or OUTGOING.
        at: (optional) a specific branch to find pull requests to or from.
        state: (optional, defaults to OPEN) only pull requests in the specified state will be returned.
            Either OPEN, DECLINED or MERGED.
        order: (optional) the order to return pull requests in, either OLDEST (as in: "oldest first") or NEWEST.
        """

        params = {}

        if direction is not None:
            params['direction'] = direction
        if at is not None:
            params['at'] = at
        if state is not None:
            params['state'] = state
        if order is not None:
            params['order'] = order
        if author is not None:
            params['role.1'] = 'AUTHOR'
            params['username.1'] = author

        return self.paginate("", params=params)

    def _make_ref(self, ref, refName="the ref"):
        if isinstance(ref, basestring):
            repo = self._parent.get()
            return PullRequestRef(repo['project']['key'], repo['slug'], ref).to_dict()
        elif isinstance(ref, PullRequestRef):
            return ref.to_dict()
        elif isinstance(ref, dict):
            return ref
        else:
            raise ValueError(refName + " should be either a string, a dict, or a PullRequestRef")

    @response_or_error
    def create(self, title, fromRef, toRef, description='', state='OPEN', reviewers=None):
        """
        Create a new pull request between two branches.
        """
        data = dict(title=title,
                    description=description,
                    fromRef=self._make_ref(fromRef, "fromRef"),
                    toRef=self._make_ref(toRef, "toRef"),
                    state=state)

        if reviewers is not None:
            data['reviewers'] = []
            for reviewer in reviewers:
                data['reviewers'].append({"user": dict(name=reviewer)})

        return self._client.post(self.url(""), data=data)

    def __getitem__(self, item):
        """
        Return a specific pull requests
        """
        return PullRequest(item, self.url(str(item)), self._client, self)

    @response_or_error
    def get(self):
        """
        Retrieve the settings for a pull requests workflow
        """
        return self._client.get(self.url())

    @response_or_error
    def configure(self, configuration=None):
        """
        Modify the settings for a pull requests workflow
        """
        return self._client.post(self.url(), data=configuration)





import sys

_ver = sys.version_info

#: Python 2.x?
is_py2 = (_ver[0] == 2)

#: Python 3.x?
is_py3 = (_ver[0] == 3)


if is_py2:
    def update_doc(method, newdoc):
        method.im_func.func_doc = newdoc

    basestring = basestring
elif is_py3:
    def update_doc(method, newdoc):
        method.__doc__ = newdoc

    basestring = (str, bytes)



"""
Stashy Build Status Notifications

"""

from enum import Enum

from stashy.helpers import ResourceBase

from stashy.errors import ok_or_error
from stashy.errors import response_or_error


API_NAME = "build-status"
API_VERSION = "1.0"
API_OVERRIDE_PATH = "{api}/{version}/commits".format


class BuildStates(Enum):
    SUCCESSFUL = "SUCCESSFUL"
    FAILED = "FAILED"
    INPROGRESS = "INPROGRESS"


class Build(ResourceBase):
    def __init__(self, url, client, git_hash):
        super(Build, self).__init__(url, client, None, API_OVERRIDE_PATH(api=API_NAME, version=API_VERSION))
        self._git_hash = git_hash

    def url(self, *args, **kwargs):
        return super(Build, self).url(self._git_hash)

    @ok_or_error
    def set(self, state, key, name, url, description=None):
        """
        Set a build notification for this hash.
           https://developer.atlassian.com/bitbucket/server/docs/latest/how-tos/updating-build-status-for-commits.html
       """
        description = key if description is None else description

        state = BuildStates(state)

        data = {"state": state.value,
                "key": key,
                "name": name,
                "url": url,
                "description": description}

        return self._client.post(self.url(), data=data)

    @response_or_error
    def get(self):
        """
        Get all of the build notifications for this hash
        """
        return self._client.get(self.url())



from functools import wraps
from decorator import decorator

class NotFoundException(Exception):
    def __init__(self, response):
        try:
            self.data = response.json()
            if 'errors' in self.data:
                msg = self.data['errors'][0]['message']
            else:
                msg = str(self.data)
        except ValueError:
            msg = "Not found: " + response.url

        super(NotFoundException, self).__init__(msg)


class GenericException(Exception):
    def __init__(self, response):
        try:
            self.data = response.json()
            msg = "%d: %s" % (response.status_code, self.data)
        except ValueError:
            msg = "Unknown error: %d(%s)" % (response.status_code, response.reason)

        super(GenericException, self).__init__(msg)


class AuthenticationException(Exception):
    def __init__(self, response):
        try:
            msg = "%d: Invalid User / Password" % response.status_code
        except ValueError:
            msg = "Invalid Authentication" 

        super(AuthenticationException,self).__init__(msg)

def maybe_throw(response):
    if not response.ok:
        if response.status_code == 404:
            raise NotFoundException(response)
        elif response.status_code == 401:
            raise AuthenticationException(response)
        else:
            e = GenericException(response)
            try:
                e.data = response.json()
            except ValueError:
                e.content = response.content
            raise e


@decorator
def ok_or_error(fn, *args, **kw):
    response = fn(*args, **kw)
    maybe_throw(response)
    return response.ok


@decorator
def response_or_error(fn, *args, **kw):
    response = fn(*args, **kw)
    maybe_throw(response)
    try:
        return response.json()
    except ValueError:
        return response.text



import json
import requests

from .helpers import Nested, add_json_headers
from .admin import Admin
from .projects import Projects
from .ssh import Keys
from .compat import basestring
from .allrepos import Repos
from .builds import Build

class Stash(object):
    _url = "/"

    def __init__(self, base_url, username=None, password=None, oauth=None, verify=True, token=None, session=None):
        self._client = StashClient(base_url, username, password, oauth, verify, token, session=session)

    admin = Nested(Admin)
    projects = Nested(Projects)
    ssh = Nested(Keys)
    repos = Nested(Repos)

    def groups(self, filter=None):
        """
        Consider using stash.admin.groups instead.
        """
        return self.admin.groups.get(filter)

    def users(self, filter=None):
        """
        Consider using stash.admin.users instead.
        """
        return self.admin.users.get(filter)

    def build(self, git_hash):
        """
        Get build information for a commit.
        """
        return Build("", self._client, git_hash)


class StashClient(object):

    # the core api path will be used as an overridable default
    core_api_name = 'api'
    core_api_version = '1.0'
    core_api_path = '{0}/{1}'.format(core_api_name, core_api_version)

    branches_api_name = 'branch-utils'
    branches_api_version = '1.0'
    branches_api_path = '{0}/{1}'.format(branches_api_name, branches_api_version)

    def __init__(self, base_url, username=None, password=None, oauth=None, verify=True, token=None, session=None):
        assert isinstance(base_url, basestring)

        if base_url.endswith("/"):
            self._base_url = base_url[:-1]
        else:
            self._base_url = base_url

        self._api_base = self._base_url + "/rest"

        if session is None:
            session = requests.Session()

        self._session = session
        self._session.verify = verify

        if oauth is not None:
            self._create_oauth_session(oauth)
        elif username is not None or password is not None:
            self._session.auth = (username, password)
        elif token is not None:
            self._session.headers.update({'Authorization': 'Bearer {}'.format(token)})

        self._session.cookies = self._session.head(self.url("")).cookies
        self._session.headers.update({'Content-Type': 'application/json'})

    def _create_oauth_session(self, oauth):
        from requests_oauthlib import OAuth1
        from oauthlib.oauth1 import SIGNATURE_RSA

        oauth = OAuth1(
            oauth['consumer_key'],
            rsa_key=oauth['key_cert'],
            signature_method=SIGNATURE_RSA,
            resource_owner_key=oauth['access_token'],
            resource_owner_secret=oauth['access_token_secret']
        )
        self._session.auth = oauth

    def url(self, resource_path):
        assert isinstance(resource_path, basestring)
        if not resource_path.startswith("/"):
            resource_path = "/" + resource_path
        return self._api_base + resource_path

    def head(self, resource, **kw):
        return self._session.head(self.url(resource), **kw)

    def get(self, resource, **kw):
        return self._session.get(self.url(resource), **kw)

    def post(self, resource, data=None, **kw):
        if data:
            kw = add_json_headers(kw)
            data = json.dumps(data)
        return self._session.post(self.url(resource), data, **kw)

    def put(self, resource, data=None, **kw):
        if data:
            kw = add_json_headers(kw)
            data = json.dumps(data)
        return self._session.put(self.url(resource), data, **kw)

    def delete(self, resource, data=None,**kw):
        if data:
            data = json.dumps(data)
            kw = add_json_headers(kw)
            return self._session.request(method='DELETE', url=self.url(
                resource), data=data, **kw)
        return self._session.delete(self.url(resource), **kw)



# from .helpers import Nested, ResourceBase, IterableResource
from .helpers import ResourceBase, IterableResource
from .errors import ok_or_error, response_or_error
from .compat import update_doc


class SshFilteredIterableResource(IterableResource):
    def all(self, user=None):
        """
        Retrieve all the resources, optionally modified by filter.
        """
        params = {}
        if user:
            params['user'] = user
        return self.paginate("", params)

    def list(self, user=None):
        """
        Convenience method to return a list (rather than iterable) of all
        elements
        """
        return list(self.all(user))


class Key(ResourceBase):
    def __init__(self, url, client, parent):
        super(Keys, self).__init__(url, client, parent)
        self._url = 'ssh/1.0/keys'

    def get(self):
        return self._client.get(self.url())


class Keys(ResourceBase, SshFilteredIterableResource):
    def __init__(self, url, client, parent):
        super(Keys, self).__init__(url, client, parent)
        self._url = 'ssh/1.0/keys'

    @ok_or_error
    def create(self, user, key, label=None):
        """
        Adds the key for the supplied user. If label is not set then
        the comment part of the key is used as it.
        """
        data = dict(text=key, label=label)
        params = dict(user=user)
        return self._client.post(self.url(""), data=data, params=params)

    @response_or_error
    def get(self, user):
        """
        Retrieve the keys matching the supplied user.
        """
        params = dict(user=user)
        return self._client.get(self.url(""), params=params)

    def __getitem__(self, item):
        return Key(item, self.url(item), self._client, self)


update_doc(Keys.all, """Retrieve keys.""")



from .helpers import Nested, ResourceBase, IterableResource
from .errors import ok_or_error, response_or_error
from .permissions import Permissions, RepositoryPermissions
from .pullrequests import PullRequests
from .compat import update_doc
from .branch_permissions import BranchPermissions
import json

class Hook(ResourceBase):
    def __init__(self, key, url, client, parent):
        super(Hook, self).__init__(url, client, parent)
        self._key = key

    @response_or_error
    def get(self):
        """
        Retrieve a repository hook
        """
        return self._client.get(self.url())

    @response_or_error
    def enable(self, configuration=None):
        """
        Enable a repository hook, optionally applying new configuration.
        """
        return self._client.put(self.url("/enabled"), data=configuration)

    @response_or_error
    def disable(self):
        """
        Disable a repository hook
        """
        return self._client.delete(self.url("/enabled"))

    @response_or_error
    def settings(self):
        """
        Retrieve the settings for a repository hook
        """
        return self._client.get(self.url("/settings"))

    @response_or_error
    def configure(self, configuration):
        """
        Modify the settings for a repository hook
        """
        return self._client.put(self.url("/settings"), data=configuration)


class Hooks(ResourceBase, IterableResource):
    def all(self, type=None):
        """
        Retrieve hooks for this repository, optionally filtered by type.

        type: Valid values are PRE_RECEIVE or POST_RECEIVE
        """
        params=None
        if type is not None:
            params = dict(type=type)
        return self.paginate("", params=params)

    def list(self, type=None):
        """
        Convenience method to return a list (rather than iterable) of all elements
        """
        return list(self.all(type=type))

    def __getitem__(self, item):
        """
        Return a :class:`Hook` object for operations on a specific hook
        """
        return Hook(item, self.url(item), self._client, self)

class Webhook(ResourceBase):
    def __init__(self, key, url, client, parent):
        super(Webhook, self).__init__(url, client, parent)
        self._key = key

    @response_or_error
    def get(self):
        """
        Retrieve a repository webhook
        """
        return self._client.get(self.url())

    @response_or_error
    def enable(self, configuration=None):
        """
        Enable a repository webhook, optionally applying new configuration.
        """
        return self._client.put(self.url("/enabled"), data=configuration)

    @response_or_error
    def disable(self):
        """
        Disable a repository webhook
        """
        return self._client.delete(self.url("/enabled"))

    @response_or_error
    def settings(self):
        """
        Retrieve the settings for a repository webhook
        """
        return self._client.get(self.url("/settings"))

    @response_or_error
    def configure(self, configuration):
        """
        Modify the settings for a repository webhook
        """
        return self._client.put(self.url("/settings"), data=configuration)


class Webhooks(ResourceBase, IterableResource):
    def all(self, type=None):
        """
        Retrieve webhooks for this repository, optionally filtered by type.

        type: Valid values are PRE_RECEIVE or POST_RECEIVE
        """
        params=None
        if type is not None:
            params = dict(type=type)
        return self.paginate("", params=params)

    @response_or_error
    def create(self, name, url, events=["repo:refs_changed"], active=True):
        """
        Create a webhook with the given name and url
        """
        return self._client.post(self.url(), data={"name": name,
                                                   "active": active,
                                                   "events": events,
                                                   "url": url
                                                   })

    def list(self, type=None):
        """
        Convenience method to return a list (rather than iterable) of all elements
        """
        return list(self.all(type=type))

    def __getitem__(self, item):
        """
        Return a :class:`Webhook` object for operations on a specific hook
        """
        return Webhook(item, self.url(item), self._client, self)

class Settings(ResourceBase):
    hooks = Nested(Hooks)
    pullrequests = Nested(PullRequests, relative_path="/pull-requests")


class Repository(ResourceBase):
    def __init__(self, slug, url, client, parent):
        super(Repository, self).__init__(url, client, parent)
        self._slug = slug

    @response_or_error
    def delete(self):
        """
        Schedule the repository to be deleted
        """
        return self._client.delete(self.url())

    @response_or_error
    def update(self, name):
        """
        Update the name of a repository.

        The repository's slug is derived from its name. If the name changes the slug may also change.
        """
        return self._client.put(self.url(), data=dict(name=name))

    @response_or_error
    def get(self):
        """
        Retrieve the repository
        """
        return self._client.get(self.url())

    @response_or_error
    def fork(self, name = None, project = None):
        """
        Fork the repository.

        name    - Specifies the forked repository's name
                    Defaults to the name of the origin repository if not specified
        project - Specifies the forked repository's target project by key
                    Defaults to the current user's personal project if not specified
        
        """
        data = dict()
        if name is not None:
            data['name'] = name
        if project is not None:
            data['project'] = {"key": project}
            
        return self._client.post(self.url(), data=data)

    def forks(self):
        """
        Retrieve repositories which have been forked from this one.
        """
        return self.paginate('/forks')

    @response_or_error
    def tags(self, filterText=None, orderBy=None):
        """
        Retrieve the tags matching the supplied filterText param.
        """
        params = {}
        if filterText is not None:
            params['filterText'] = filterText
        if orderBy is not None:
            params['orderBy'] = orderBy
        return self._client.get(self.url('/tags'), params=params)

    @response_or_error
    def _get_default_branch(self):
        return self._client.get(self.url('/branches/default'))

    @ok_or_error
    def _set_default_branch(self, value):
        return self._client.put(self.url('/branches/default'), data=dict(id=value))

    @ok_or_error
    def create_branch(self, value, origin_branch='master'):
        return self._client.post(self.url('/branches', is_branches=True),
                                data=dict(name=value, startPoint=
                                "refs/heads/%s" % origin_branch))

    @ok_or_error
    def delete_branch(self, value):
        return self._client.delete(self.url('/branches', is_branches=True),
                                data=dict(name=value,
                                          dryRun='false'))
    @response_or_error
    def get_branch_info(self, changesetId):
        return self._client.get(self.url('/branches/info/%s' % changesetId,
                                            is_branches=True))

    def branches(self, filterText=None, orderBy=None, details=None):
        """
        Retrieve the branches matching the supplied filterText param.
        """
        params = {}
        if filterText is not None:
            params['filterText'] = filterText
        if orderBy is not None:
            params['orderBy'] = orderBy
        if details is not None:
            params['details'] = details
        return self.paginate('/branches', params=params)

    default_branch = property(_get_default_branch, _set_default_branch, doc="Get or set the default branch")

    def get_all_branches(self, items):
        """
        Return list of all branches in this project and the repository
        :param items: limit parameter (max items in result)
        :return: 
        """
        branches = self._client.get(self.url('/branches?limit={}'.format(items)))
        return json.loads(branches.content)

    def get_all_tags(self, items):
        """
        Return list of all tags in this project and the repository
        :param items: limit parameter (max items in result)
        :return:
        """
        tags = self._client.get(self.url('/tags?limit={}'.format(items)))
        return json.loads(tags.content)

    def get_commit(self, commit):
        """
        Returns detailed information about a given commit
        :param commit: like "1c972ea39318a4b3ce99bc51ab03277138c586ea"
        :return: 
        """
        res = self._client.get(self.url('/commits/{}'.format(commit)))
        return json.loads(res.content)

    def files(self, path='', at=None):
        """
        Retrieve a page of files from particular directory of a repository. The search is done
        recursively, so all files from any sub-directory of the specified directory will be returned.
        """
        params = {}
        if at is not None:
            params['at'] = at
        return self.paginate('/files/' + path, params)

    def browse(self, path='', at=None, type=False, blame='', noContent=''):
        """
        Retrieve a page of content for a file path at a specified revision.
        """
        params = {}
        if at is not None:
            params['at'] = at
        if type:
            params['type'] = type
            return response_or_error(lambda: self._client.get(self.url('/browse/' + path), params=params))()
        else:
            if blame:
                params['blame'] = blame
            if noContent:
                params['noContent'] = noContent

            return self.paginate("/browse/" + path, params=params, values_key='lines')

    def changes(self, until, since=None):
        """
        Retrieve a page of changes made in a specified commit.

        since: the changeset to which until should be compared to produce a page of changes.
               If not specified the parent of the until changeset is used.

        until: the changeset to retrieve file changes for.
        """
        params = dict(until=until)
        if since is not None:
            params['since'] = since
        return self.paginate('/changes', params=params)

    def commits(self, until, since=None, path=None):
        """
        Retrieve a page of changesets from a given starting commit or between two commits.
        The commits may be identified by hash, branch or tag name.

        since: the changeset id or ref (exclusively) to restrieve changesets after
        until: the changeset id or ref (inclusively) to retrieve changesets before.
        path: an optional path to filter changesets by.

        Support for withCounts is not implement.
        """
        params = dict(until=until, withCounts=False)
        if since is not None:
            params['since'] = since
        if path is not None:
            params['path'] = path
        return self.paginate('/commits', params=params)

    permissions = Nested(Permissions)
    repo_permissions = Nested(RepositoryPermissions,
                              relative_path="/permissions")

    pull_requests = Nested(PullRequests, relative_path="/pull-requests")
    settings = Nested(Settings)
    webhooks = Nested(Webhooks)
    branch_permissions = Nested(BranchPermissions, relative_path=None)


class Repos(ResourceBase, IterableResource):
    @response_or_error
    def create(self, name, scmId="git", forkable=True):
        """
        Create a repository with the given name
        """
        return self._client.post(self.url(), data={"name": name,
                                                   "scmId": scmId,
                                                   "forkable": forkable,
                                                   })

    def __getitem__(self, item):
        """
        Return a :class:`Repository` object for operations on a specific repository
        """
        return Repository(item, self.url(item), self._client, self)


update_doc(Repos.all, """Retrieve repositories from the project""")



from .errors import maybe_throw


def add_json_headers(kw):
    if 'headers' not in kw:
        kw['headers'] = {}

    headers = {'Content-type': 'application/json', 'Accept': 'application/json'}

    for header, value in headers.items():
        kw['headers'][header] = value

    return kw


class ResourceBase(object):
    def __init__(self, url, client, parent, api_path=None,
                 branches_api_path=None):
        self._client = client
        self._parent = parent
        if api_path is None:
            api_path = self._client.core_api_path
            branches_api_path = self._client.branches_api_path
        if branches_api_path is None:
            branches_api_path = self._client.branches_api_path

        # make sure we're only prefixing with one api path
        if url.startswith(api_path):
            self._url = url
            self._branchesurl = url.replace(api_path, branches_api_path)
        elif url.startswith(self._client.core_api_path):
            self._url = url.replace(self._client.core_api_path, api_path)
            self._branchesurl = url.replace(self._client.core_api_path,
                                            branches_api_path)
        else:
            if url.startswith('/'):
                url = url[1:]
            self._url = '{0}/{1}'.format(api_path, url)
            self._branchesurl = '{0}/{1}'.format(branches_api_path, url)




    def url(self, resource_url="", is_branches=False):
        if resource_url and not resource_url.startswith("/"):
            resource_url = "/" + resource_url
        if is_branches:
            if self._url.endswith("/"):
                url = self._branchesurl[:-1]
            else :
                url = self._branchesurl
        else:
            if self._url.endswith("/"):
                url = self._url[:-1]
            else :
                url = self._url
        return url + resource_url

    def paginate(self, resource_url, params=None, values_key='values',
                 is_branches=False):
        url = self.url(resource_url, is_branches)

        more = True
        start = None
        while more:
            kw = {}
            if params:
                kw['params'] = params
            if start is not None:
                kw.setdefault('params', {})
                kw['params']['start'] = start

            response = self._client.get(url, **kw)
            maybe_throw(response)

            data = response.json()

            if not values_key in data:
                return
            for item in data[values_key]:
                yield item

            if data['isLastPage']:
                more = False
            else:
                more = True
                start = data['nextPageStart']


class IterableResource(object):
    def __iter__(self):
        """
        Convenience method around self.all()
        """
        return self.all()

    def all(self):
        """
        Retrieve all the resources.
        """
        return self.paginate("")

    def list(self):
        """
        Convenience method to return a list (rather than iterable) of all elements
        """
        return list(self.all())


class FilteredIterableResource(IterableResource):
    def all(self, filter=None):
        """
        Retrieve all the resources, optionally modified by filter.
        """
        params = {}
        if filter:
            params['filter'] = filter
        return self.paginate("", params)

    def list(self, filter=None):
        """
        Convenience method to return a list (rather than iterable) of all elements
        """
        return list(self.all(filter))


class Nested(object):
    def __init__(self, cls, relative_path=''):

        # nested object for clarity of usage, no effect on resource url
        if relative_path is None:
            self.relative_path = ''

        # default case, use lowercase class name
        elif not relative_path:
            self.relative_path = "/%s" % cls.__name__.lower()

        # explicit override of relative path
        else:
            if not relative_path.startswith("/"):
                relative_path = "/" + relative_path
            self.relative_path = relative_path
        self.cls = cls

    def __get__(self, instance, kind):
        parent_url = instance._url
        if parent_url.endswith("/"):
            parent_url = parent_url[:-1]

        url = parent_url + self.relative_path
        return self.cls(url=url, client=instance._client, parent=instance)




from .helpers import ResourceBase, IterableResource
from .repos import Repository
from .compat import update_doc

class Repos(ResourceBase, IterableResource):
  def __getitem__(self, item):
    """
    Return a :class:`Repository` object for operations on a specific repository
    """
    return Repository(item, self.url(item), self._client, self)

update_doc(Repos.all, """Retreive repositories from Stash""")



__version__ = "0.6"

from .client import Stash

def connect(url, username, password, verify=True):
    """Connect to a Stash instance given a username and password.

    This is only recommended via SSL. If you need are using
    self-signed certificates, you can use verify=False to ignore SSL
    verifcation.
    """
    return Stash(url, username, password, verify=verify)

__all__ = ['connect']



from .helpers import ResourceBase, Nested, FilteredIterableResource
from .errors import ok_or_error
from .compat import update_doc

class Groups(ResourceBase, FilteredIterableResource):
    def none(self, filter=None):
        """
        Retrieve groups that have no granted permissions.

        filter: return only group names containing the supplied string will be returned
        """
        params = {}
        if filter:
            params['filter'] = filter
        return self.paginate("/none", params)

    @ok_or_error
    def grant(self, group, permission):
        """
        Promote or demote a user's permission level.

        Depending on context, you may use one of the following set of permissions:

        global permissions:

            * LICENSED_USER
            * PROJECT_CREATE
            * ADMIN
            * SYS_ADMIN

        project permissions:
            * PROJECT_READ
            * PROJECT_WRITE
            * PROJECT_ADMIN
        """
        return self._client.put(self.url(), params=dict(name=group, permission=permission))

    @ok_or_error
    def revoke(self, group):
        """
        Revoke all permissions for a group.
        """
        return self._client.delete(self.url(), params=dict(name=group))


class Users(ResourceBase, FilteredIterableResource):
    def none(self, filter=None):
        """
        Retrieve users that have no granted permissions.

        filter: if specified only user names containing the supplied string will be returned
        """
        params = {}
        if filter:
            params['filter'] = filter
        return self.paginate("/none", params)

    @ok_or_error
    def grant(self, user, permission):
        """
        Promote or demote the permission level of a user.


        Depending on context, you may use one of the following set of permissions:

        global permissions:

            * LICENSED_USER
            * PROJECT_CREATE
            * ADMIN
            * SYS_ADMIN

        project permissions:
            * PROJECT_READ
            * PROJECT_WRITE
            * PROJECT_ADMIN

        repository permissions:
            * REPO_READ
            * REPO_WRITE
            * REPO_ADMIN

        """
        return self._client.put(self.url(), params=dict(name=user, permission=permission))

    @ok_or_error
    def revoke(self, user):
        """
        Revoke all permissions for a user.
        """
        return self._client.delete(self.url(), params=dict(name=user))


class Permissions(ResourceBase):
    groups = Nested(Groups)
    users = Nested(Users)


class ProjectPermissions(Permissions):
    def _url_for(self, permission):
        return self.url().rstrip("/") + "/" + permission + "/all"

    @ok_or_error
    def grant(self, permission):
        """
        Grant or revoke a project permission to all users, i.e. set the default permission.

        project permissions:
            * PROJECT_READ
            * PROJECT_WRITE
            * PROJECT_ADMIN

        """
        return self._client.post(self._url_for(permission), params=dict(allow=True))

    @ok_or_error
    def revoke(self, permission):
        """
        Revoke a project permission from all users, i.e. revoke the default permission.

        project permissions:
            * PROJECT_READ
            * PROJECT_WRITE
            * PROJECT_ADMIN

        """
        return self._client.post(self._url_for(permission), params=dict(allow=False))

class RepositoryPermissions(Permissions):
    def _url_for(self):
        return self.url().rstrip("/") + "/users"

    @ok_or_error
    def grant(self, user, permission):
        """
        Grant or revoke a repository permission to all users, i.e. set the
        default permission.

        repository permissions:
            * REPO_READ
            * REPO_WRITE
            * REPO_ADMIN


        """
        return self._client.put(self._url_for(), params=dict(name=user,
                                                              permission=permission))

    @ok_or_error
    def revoke(self, user):
        """
        Revoke a repository permission from all users, i.e. revoke the default
        permission.

        repository permissions:
            * REPO_READ
            * REPO_WRITE
            * REPO_ADMIN

        """
        return self._client.post(self._url_for(), params=dict(name=user))


update_doc(Groups.all, """
Returns groups that have been granted at least one permission.

filter: return only group names containing the supplied string will be returned
""")

update_doc(Users.all, """
Returns users that have been granted at least one permission.

filter: if specified only user names containing the supplied string will be returned
""")
